@@ -27,11 +27,15 @@
 
 #include <libgossip/gossip-utils.h>
 
-#include "gossip-jid.h"
-#include "gossip-jabber-helper.h"
 #include "gossip-jabber-chatrooms.h"
+#include "gossip-jid.h"
+#include "gossip-jabber-utils.h"
+#include "gossip-jabber-private.h"
 
-#define d(x) x 
+#define DEBUG_MSG(x) 
+/* #define DEBUG_MSG(args) g_printerr args ; g_printerr ("\n"); */
+
+#define JOIN_TIMEOUT       20000
 
 #define JOIN_MSG_ID_PREFIX "gc_join_"
 
@@ -47,109 +51,232 @@ struct _GossipJabberChatrooms {
 
 
 typedef struct {
-	gint              id;
-	GossipJID        *jid;
-	gchar            *name;
+	gint                   ref_count;
 
-	GossipContact    *own_contact;
-	LmMessageHandler *join_handler;
-	
-	GSList           *contacts;
+	GossipChatroom        *chatroom;
+
+	GossipJID             *jid;
+	GossipContact         *own_contact;
+
+	GSList                *contacts;
+
+	guint                  timeout_id;
+
+ 	GossipJabberChatrooms *chatrooms; 
+        GossipChatroomJoinCb   callback;
+        gpointer               user_data;
+
+	LmConnection          *connection;
+	LmMessageHandler      *join_handler;
 } JabberChatroom;
 
 
-typedef struct {
+static JabberChatroom * jabber_chatrooms_chatroom_new         (GossipJabberChatrooms  *chatrooms,
+							       GossipChatroom         *chatroom);
+static JabberChatroom * jabber_chatrooms_chatroom_ref         (JabberChatroom         *room);
+static void             jabber_chatrooms_chatroom_unref       (JabberChatroom         *room);
+static GossipChatroomId jabber_chatrooms_chatroom_get_id      (JabberChatroom         *room);
+static LmHandlerResult  jabber_chatrooms_message_handler      (LmMessageHandler       *handler,
+							       LmConnection           *conn,
+							       LmMessage              *message,
+							       GossipJabberChatrooms  *chatrooms);
+static LmHandlerResult  jabber_chatrooms_presence_handler     (LmMessageHandler       *handler,
+							       LmConnection           *conn,
+							       LmMessage              *message,
+							       GossipJabberChatrooms  *chatrooms);
+static LmHandlerResult  jabber_chatrooms_join_cb              (LmMessageHandler       *handler,
+							       LmConnection           *connection,
+							       LmMessage              *message,
+							       JabberChatroom         *room);
+static GossipContact *  jabber_chatrooms_get_contact          (JabberChatroom         *room,
+							       GossipJID              *jid,
+							       gboolean               *new_contact);
+static void             jabber_chatrooms_get_rooms_foreach    (gpointer                key,
+							       JabberChatroom         *room,
+							       GList                 **list);
+static void             jabber_chatrooms_set_presence_foreach (gpointer                key,
+							       JabberChatroom         *room,
+							       GossipJabberChatrooms  *chatrooms);
+
+
+GossipJabberChatrooms *
+gossip_jabber_chatrooms_init (GossipJabber *jabber)
+{
 	GossipJabberChatrooms *chatrooms;
-        gpointer               callback;
-        gpointer               user_data;
-} AsyncCallbackData;
+	LmConnection          *connection;
+	LmMessageHandler      *handler;
 
+	g_return_val_if_fail (GOSSIP_IS_JABBER (jabber), NULL);
 
-static JabberChatroom *jabber_chatrooms_chatroom_new         (const gchar            *room_name,
-							      const gchar            *server,
-							      const gchar            *nick);
-static void            jabber_chatrooms_chatroom_free        (JabberChatroom         *room);
-static LmHandlerResult jabber_chatrooms_message_handler      (LmMessageHandler       *handler,
-							      LmConnection           *conn,
-							      LmMessage              *message,
-							      GossipJabberChatrooms  *chatrooms);
-static LmHandlerResult jabber_chatrooms_presence_handler     (LmMessageHandler       *handler,
-							      LmConnection           *conn,
-							      LmMessage              *message,
-							      GossipJabberChatrooms  *chatrooms);
-static LmHandlerResult jabber_chatrooms_join_cb              (LmMessageHandler       *handler,
-							      LmConnection           *connection,
-							      LmMessage              *message,
-							      AsyncCallbackData      *data);
-static GossipContact * jabber_chatrooms_get_contact          (JabberChatroom         *room,
-							      GossipJID              *jid,
-							      gboolean               *new_contact);
-static void            jabber_chatrooms_get_rooms_foreach    (gpointer                key,
-							      JabberChatroom         *room,
-							      GList                 **list);
-static void            jabber_chatrooms_set_presence_foreach (gpointer                key,
-							      JabberChatroom         *room,
-							      GossipJabberChatrooms  *chatrooms);
+	connection = gossip_jabber_get_connection (jabber);
+	g_return_val_if_fail (connection != NULL, NULL);
 
+	chatrooms = g_new0 (GossipJabberChatrooms, 1);
+	
+	chatrooms->jabber     = g_object_ref (jabber);
+	chatrooms->connection = lm_connection_ref (connection);
+	chatrooms->presence   = NULL;
+	chatrooms->room_id_hash = g_hash_table_new (NULL, NULL);
+	chatrooms->room_jid_hash = g_hash_table_new_full (gossip_jid_hash,
+							  gossip_jid_equal,
+							  (GDestroyNotify) gossip_jid_unref,
+							  NULL);
+
+	handler = lm_message_handler_new ((LmHandleMessageFunction) jabber_chatrooms_message_handler,
+					  chatrooms, NULL);
+
+	lm_connection_register_message_handler (chatrooms->connection,
+						handler, 
+						LM_MESSAGE_TYPE_MESSAGE,
+						LM_HANDLER_PRIORITY_NORMAL);
+	lm_message_handler_unref (handler);
+	
+	handler = lm_message_handler_new ((LmHandleMessageFunction) jabber_chatrooms_presence_handler,
+					  chatrooms, NULL);
+
+	lm_connection_register_message_handler (chatrooms->connection,
+						handler, 
+						LM_MESSAGE_TYPE_PRESENCE,
+						LM_HANDLER_PRIORITY_FIRST);
+	lm_message_handler_unref (handler);
+
+	return chatrooms;
+}	
+
+void
+gossip_jabber_chatrooms_finalize (GossipJabberChatrooms *chatrooms)
+{
+	if (!chatrooms) {
+		/* We don't error here, because if no connection is
+		made, then we can clean up a GossipJabber object
+		without any chatrooms ever existing */
+		return;
+	}
+
+	g_hash_table_destroy (chatrooms->room_id_hash);
+	g_hash_table_destroy (chatrooms->room_jid_hash);
+
+	if (chatrooms->presence) {
+		g_object_unref (chatrooms->presence);
+	}
+
+	lm_connection_unref (chatrooms->connection);
+	g_object_unref (chatrooms->jabber);
+	
+	g_free (chatrooms);
+}
 
 static JabberChatroom *
-jabber_chatrooms_chatroom_new (const gchar *room_name, 
-			       const gchar *server,
-			       const gchar *nick)
+jabber_chatrooms_chatroom_new (GossipJabberChatrooms *chatrooms,
+			       GossipChatroom        *chatroom)
 {
+	GossipJabber   *jabber;
+	GossipContact  *own_contact;
 	JabberChatroom *room;
 	gchar          *jid_str;
-	static int      id = 1;
+
+	jabber = chatrooms->jabber;
 
 	room = g_new0 (JabberChatroom, 1);
 
-	room->id = id++;
-	room->name = g_strdup (room_name);
+	room->ref_count = 1;
+	room->chatroom = g_object_ref (chatroom);
 
-	jid_str = g_strdup_printf ("%s@%s/%s", room_name, server, nick);
+	jid_str = g_strdup_printf ("%s@%s/%s", 
+				   gossip_chatroom_get_room (chatroom), 
+				   gossip_chatroom_get_server (chatroom),
+				   gossip_chatroom_get_nick (chatroom));
 	room->jid = gossip_jid_new (jid_str);
 	g_free (jid_str);
 
 	room->contacts = NULL;
-	
-	room->own_contact = gossip_contact_new_full (GOSSIP_CONTACT_TYPE_USER,
-						     gossip_jid_get_full (room->jid),
-						     gossip_jid_get_resource (room->jid));
+
+	/* What we do here is copy the contact instead of reference
+	 * it, plus we change the name according to how the user wants
+	 * their nickname in the chat room. 
+	 */
+	own_contact = gossip_jabber_get_own_contact (jabber);
+	room->own_contact = gossip_contact_copy (own_contact);
+	gossip_contact_set_name (room->own_contact, 
+				 gossip_chatroom_get_nick (chatroom));
 
 	return room;
 }
 
-static void
-jabber_chatrooms_chatroom_free (JabberChatroom *room)
+static JabberChatroom *
+jabber_chatrooms_chatroom_ref (JabberChatroom *room)
 {
-	GSList *l;
+	if (!room) {
+		return NULL;
+	}
 	
-	g_free (room->name);
-	gossip_jid_unref (room->jid);
+	room->ref_count++;
+	return room;
+}
 
-	for (l = room->contacts; l; l = l->next) {
-		g_object_unref (l->data);
+static void
+jabber_chatrooms_chatroom_unref (JabberChatroom *room)
+{
+	if (!room) {
+		return;
 	}
 
-	g_object_unref (room->own_contact);
+	room->ref_count--;
+
+	if (room->ref_count > 0) {
+		return;
+	}
+
+	g_object_unref (room->chatroom);
+
+	gossip_jid_unref (room->jid);
+
+	g_slist_foreach (room->contacts, (GFunc)g_object_unref, NULL);
 	g_slist_free (room->contacts);
 
-	lm_message_handler_unref (room->join_handler);
+	g_object_unref (room->own_contact);
+
+	if (room->timeout_id) {
+		g_source_remove (room->timeout_id);
+		room->timeout_id = 0;
+	}
+
+	if (room->join_handler) {
+		lm_connection_unregister_message_handler (room->connection, 
+							  room->join_handler, 
+							  LM_MESSAGE_TYPE_PRESENCE);	
+		room->join_handler = NULL;
+	}
+
+	if (room->connection) {
+		lm_connection_unref (room->connection);
+		room->connection = NULL;
+	}
 
 	g_free (room);
 }
 
+static GossipChatroomId 
+jabber_chatrooms_chatroom_get_id (JabberChatroom *room)
+{
+	g_return_val_if_fail (room != NULL, 0);
+	g_return_val_if_fail (GOSSIP_IS_CHATROOM (room->chatroom), 0);
+
+	return gossip_chatroom_get_id (room->chatroom);
+}
+
 static LmHandlerResult
 jabber_chatrooms_message_handler (LmMessageHandler      *handler,
 				  LmConnection          *conn,
 				  LmMessage             *m,
 				  GossipJabberChatrooms *chatrooms)
 {
-	const gchar    *from;
-	GossipJID      *jid;
-	JabberChatroom *room;
-	GossipMessage  *message;
-	LmMessageNode  *node;
+	GossipJID        *jid;
+	GossipMessage    *message;
+	GossipChatroomId  id;
+	JabberChatroom   *room;
+	LmMessageNode    *node;
+	const gchar      *from;
 
 	if (lm_message_get_sub_type (m) != LM_MESSAGE_SUB_TYPE_GROUPCHAT) {
 		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
@@ -159,7 +286,8 @@ jabber_chatrooms_message_handler (LmMessageHandler      *handler,
 	jid = gossip_jid_new (from);
 
 	room = g_hash_table_lookup (chatrooms->room_jid_hash, jid);
-	
+	id = jabber_chatrooms_chatroom_get_id (room);
+
 	if (!room) {
 		gossip_jid_unref (jid);
 		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
@@ -170,19 +298,19 @@ jabber_chatrooms_message_handler (LmMessageHandler      *handler,
 		if (gossip_jid_get_resource (jid) == NULL) {
 			g_signal_emit_by_name (chatrooms->jabber,
 					       "chatroom-new-room-event",
-					       room->id, node->value);
+					       id, node->value);
 		} else {
 			gossip_time_t  timestamp;
 			GossipContact *contact;
 
 			contact = jabber_chatrooms_get_contact (room, jid, 
 								NULL);
-			timestamp = gossip_jabber_helper_get_timestamp_from_lm_message (m);
+			timestamp = gossip_jabber_get_message_timestamp (m);
 
 			message = gossip_message_new (GOSSIP_MESSAGE_TYPE_CHAT_ROOM,
 						      room->own_contact);
 			
-			timestamp = gossip_jabber_helper_get_timestamp_from_lm_message (m);
+			timestamp = gossip_jabber_get_message_timestamp (m);
 			gossip_message_set_timestamp (message, timestamp);
 			
 			gossip_message_set_sender (message, contact);
@@ -190,7 +318,7 @@ jabber_chatrooms_message_handler (LmMessageHandler      *handler,
 			
 			g_signal_emit_by_name (chatrooms->jabber,
 					       "chatroom-new-message", 
-					       room->id, message);
+					       id, message);
 			
 			g_object_unref (message);
 		}
@@ -200,7 +328,7 @@ jabber_chatrooms_message_handler (LmMessageHandler      *handler,
 	if (node) {
 		g_signal_emit_by_name (chatrooms->jabber,
 				       "chatroom-title-changed", 
-				       room->id, node->value);
+				       id, node->value);
 	}
 	
 	gossip_jid_unref (jid);
@@ -221,6 +349,7 @@ jabber_chatrooms_presence_handler (LmMessageHandler      *handler,
 	GossipContact       *contact;
 	GossipPresence      *presence;
 	GossipPresenceState  p_state;
+	GossipChatroomId     id;
 	LmMessageSubType     type;
 	LmMessageNode       *node;
 	gboolean             new_contact;
@@ -235,7 +364,9 @@ jabber_chatrooms_presence_handler (LmMessageHandler      *handler,
 		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 	}
 	
-	d(g_print ("Chatrooms: Presence from: %s\n", from));
+	id = jabber_chatrooms_chatroom_get_id (room);
+
+	DEBUG_MSG (("ProtocolChatrooms: Presence from: %s", from));
 
 	type = lm_message_get_sub_type (m);
 	switch (type) {
@@ -261,39 +392,65 @@ jabber_chatrooms_presence_handler (LmMessageHandler      *handler,
 
 		/* is contact new or updated */
 		if (new_contact || was_offline) {
-			d(g_print ("Chatrooms: ID[%d] Presence for new joining contact:'%s'\n",
-				   room->id,
-				   gossip_jid_get_full (jid)));
+			DEBUG_MSG (("ProtocolChatrooms: ID[%d] Presence for new joining contact:'%s'",
+				   id, gossip_jid_get_full (jid)));
 			g_signal_emit_by_name (chatrooms->jabber,
 					       "chatroom-contact-joined",
-					       room->id, contact);
+					       id, contact);
 		} else {
-			d(g_print ("Chatrooms: ID[%d] Presence updated for contact:'%s'\n", 
-				   room->id,
-				   gossip_jid_get_full (jid)));
+			DEBUG_MSG (("ProtocolChatrooms: ID[%d] Presence updated for contact:'%s'", 
+				   id, gossip_jid_get_full (jid)));
 			g_signal_emit_by_name (chatrooms->jabber,
 					       "chatroom-contact-presence-updated", 
-					       room->id, contact);
+					       id, contact);
 		}
 		break;
 
 	case LM_MESSAGE_SUB_TYPE_UNAVAILABLE:
 		contact = jabber_chatrooms_get_contact (room, jid, NULL);
 		if (contact) {
-			d(g_print ("Chatrooms: ID[%d] Contact left:'%s'\n",
-				   room->id,
-				   gossip_jid_get_full (jid)));
+			DEBUG_MSG (("ProtocolChatrooms: ID[%d] Contact left:'%s'",
+				   id, gossip_jid_get_full (jid)));
 			g_signal_emit_by_name (chatrooms->jabber, 
 					       "chatroom-contact-left",
-					       room->id, contact);
-			room->contacts = g_slist_remove (room->contacts,
-							 contact);
+					       id, contact);
+			room->contacts = g_slist_remove (room->contacts, contact);
 			g_object_unref (contact);
 		}
 		break;
 
+#if 0
+	case LM_MESSAGE_SUB_TYPE_ERROR: 
+		if (room->callback) {
+			GossipChatroomStatus status;
+
+			status = gossip_chatroom_get_status (room->chatroom);
+
+			if (status == GOSSIP_CHATROOM_CONNECTING) {
+				/* malformed join group chat message is most likely */
+				gossip_chatroom_set_status (room->chatroom, 
+							    GOSSIP_CHATROOM_ERROR);
+
+				/* clean up timeout */
+				if (room->timeout_id) {
+					g_source_remove (room->timeout_id);
+					room->timeout_id = 0;
+				}
+
+				(room->callback) (GOSSIP_CHATROOM_PROVIDER (chatrooms->jabber),
+						  GOSSIP_CHATROOM_JOIN_UNKNOWN_ERROR, 
+						  id, room->user_data);
+
+				/* clean up callback */
+				room->callback = NULL;
+				room->user_data = NULL;
+			}
+		}
+		break;
+#endif
+
 	default:
-		d(g_print ("Chatrooms: Presence not handled for:'%s'\n",
+		DEBUG_MSG (("ProtocolChatrooms: Presence not handled for:'%s'",
 			   gossip_jid_get_full (jid)));
 		gossip_jid_unref (jid);
 		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
@@ -335,126 +492,153 @@ jabber_chatrooms_get_contact (JabberChatroom *room,
 		*new_contact = TRUE;
 	}
 
-	c = gossip_contact_new_full (GOSSIP_CONTACT_TYPE_CHATROOM, id, 
+	c = gossip_contact_new_full (GOSSIP_CONTACT_TYPE_CHATROOM, 
+				     gossip_contact_get_account (room->own_contact),
+				     id, 
 				     gossip_jid_get_resource (jid));
 	room->contacts = g_slist_prepend (room->contacts, c);
 
 	return c;
 }
 
-GossipJabberChatrooms *
-gossip_jabber_chatrooms_new (GossipJabber   *jabber,
-			     LmConnection   *connection)
+static gboolean
+jabber_chatrooms_join_timeout_cb (JabberChatroom *room)
 {
+	GossipChatroomId       id;
 	GossipJabberChatrooms *chatrooms;
-	LmMessageHandler      *handler;
+	const gchar           *last_error;
 
-	chatrooms = g_new0 (GossipJabberChatrooms, 1);
-	
-	chatrooms->jabber     = jabber;
-	chatrooms->connection = connection;
-	chatrooms->presence   = NULL;
-	chatrooms->room_id_hash = g_hash_table_new (NULL, NULL);
-	chatrooms->room_jid_hash = g_hash_table_new_full (gossip_jid_hash,
-							  gossip_jid_equal,
-							  (GDestroyNotify) gossip_jid_unref,
-							  NULL);
+	room->timeout_id = 0;
 
-	handler = lm_message_handler_new ((LmHandleMessageFunction) jabber_chatrooms_message_handler,
-					  chatrooms, NULL);
-
-	lm_connection_register_message_handler (connection,
-						handler, 
-						LM_MESSAGE_TYPE_MESSAGE,
-						LM_HANDLER_PRIORITY_NORMAL);
-	lm_message_handler_unref (handler);
-	
-	handler = lm_message_handler_new ((LmHandleMessageFunction) jabber_chatrooms_presence_handler,
-					  chatrooms, NULL);
-
-	lm_connection_register_message_handler (connection,
-						handler, 
-						LM_MESSAGE_TYPE_PRESENCE,
-						LM_HANDLER_PRIORITY_FIRST);
-	lm_message_handler_unref (handler);
-
-	return chatrooms;
-}	
-
-void
-gossip_jabber_chatrooms_free (GossipJabberChatrooms *chatrooms)
-{
-	g_hash_table_destroy (chatrooms->room_id_hash);
-	g_hash_table_destroy (chatrooms->room_jid_hash);
-
-	if (chatrooms->presence) {
-		g_object_unref (chatrooms->presence);
+	if (room->join_handler) {
+		lm_message_handler_unref (room->join_handler);
+		room->join_handler = NULL;
 	}
-	
-	g_free (chatrooms);
+
+	id = jabber_chatrooms_chatroom_get_id (room);
+	DEBUG_MSG (("ProtocolChatrooms: ID[%d] Join timed out (internally)", id));
+
+	/* set chatroom status and error */
+	gossip_chatroom_set_status (room->chatroom, GOSSIP_CHATROOM_STATUS_ERROR);
+
+	last_error = gossip_chatroom_provider_join_result_as_str (GOSSIP_CHATROOM_JOIN_TIMED_OUT);
+	gossip_chatroom_set_last_error (room->chatroom, last_error);
+
+	/* call callback */
+	chatrooms = room->chatrooms;
+		
+	if (room->callback != NULL) {
+		DEBUG_MSG (("ProtocolChatrooms: ID[%d] Calling back...", id));
+		(room->callback) (GOSSIP_CHATROOM_PROVIDER (chatrooms->jabber),
+				  GOSSIP_CHATROOM_JOIN_TIMED_OUT, 
+				  id, room->user_data);
+	}
+
+	/* clean up */
+	g_hash_table_remove (chatrooms->room_id_hash, 
+			     GINT_TO_POINTER (id));
+	g_hash_table_remove (chatrooms->room_jid_hash, 
+			     room->jid);
+
+	/* clean up callback data */
+	room->callback = NULL;
+	room->user_data = NULL;
+
+	jabber_chatrooms_chatroom_unref (room);
+
+	return FALSE;
 }
 
 GossipChatroomId
 gossip_jabber_chatrooms_join (GossipJabberChatrooms *chatrooms,
-			      const gchar           *room_name,
-			      const gchar           *server,
-			      const gchar           *nick,
-			      const gchar           *password,
+			      GossipChatroom        *chatroom,
 			      GossipChatroomJoinCb   callback,
 			      gpointer               user_data)
 {
+	GossipChatroomId   id;
+	JabberChatroom    *room, *existing_room;
 	LmMessage         *m;
 	const gchar       *show = NULL;
-        AsyncCallbackData *data;
         gchar             *id_str;
-	JabberChatroom    *room, *existing_room;
 	
-        d(g_print ("Chatrooms: Join chat room:'%s' on server:'%s'\n", 
-		   room_name, server));
+	g_return_val_if_fail (chatrooms != NULL, 0);
+	g_return_val_if_fail (GOSSIP_IS_CHATROOM (chatroom), 0);
+	g_return_val_if_fail (callback != NULL, 0);
 
-	room = jabber_chatrooms_chatroom_new (room_name, server, nick);
+	room = jabber_chatrooms_chatroom_new (chatrooms, chatroom);
 	existing_room = g_hash_table_lookup (chatrooms->room_jid_hash, room->jid);
 
 	if (existing_room) {
- 		jabber_chatrooms_chatroom_free (room); 
+ 		jabber_chatrooms_chatroom_unref (room); 
 
 		/* duplicate room already exists */
+		id = jabber_chatrooms_chatroom_get_id (existing_room);
+
+		DEBUG_MSG (("ProtocolChatrooms: ID[%d] Join chatroom:'%s', room already exists.", 
+			   id,
+			   gossip_chatroom_get_room (chatroom)));
+
 		(callback) (GOSSIP_CHATROOM_PROVIDER (chatrooms->jabber),
 			    GOSSIP_CHATROOM_JOIN_ALREADY_OPEN,
-			    existing_room->id,
+			    id,
 			    user_data); 
-		
-		return 0;
+ 		
+		return id;
 	}
 
+	/* get real chatroom */
+	id = jabber_chatrooms_chatroom_get_id (room);
+
+        DEBUG_MSG (("ProtocolChatrooms: ID[%d] Join chatroom:'%s' on server:'%s'", 
+		   id,
+		   gossip_chatroom_get_room (chatroom),
+		   gossip_chatroom_get_server (chatroom)));
+
+
+	/* add timeout for server response */
+	room->timeout_id = g_timeout_add (JOIN_TIMEOUT, 
+					  (GSourceFunc) jabber_chatrooms_join_timeout_cb, 
+					  room);
+
+	room->chatrooms = chatrooms;
+
+	/* set callback data */
+        room->callback = callback;
+        room->user_data = user_data;
+	
+	gossip_chatroom_set_status (chatroom, GOSSIP_CHATROOM_STATUS_JOINING);
+	gossip_chatroom_set_last_error (room->chatroom, NULL);
+
+	/* compose message */
 	m = lm_message_new_with_sub_type (gossip_jid_get_full (room->jid),
 					  LM_MESSAGE_TYPE_PRESENCE,
-					  
 					  LM_MESSAGE_SUB_TYPE_AVAILABLE);
 
 	g_hash_table_insert (chatrooms->room_id_hash, 
-			     GINT_TO_POINTER (room->id), room);
-	g_hash_table_insert (chatrooms->room_jid_hash, room->jid, room);
+			     GINT_TO_POINTER (id), 
+			     jabber_chatrooms_chatroom_ref (room));
+	g_hash_table_insert (chatrooms->room_jid_hash, 
+			     room->jid, 
+			     jabber_chatrooms_chatroom_ref (room));
 
-        show = gossip_jabber_helper_presence_state_to_string (chatrooms->presence);
+	/* release the local ref */
+	jabber_chatrooms_chatroom_unref (room); 
+
+        show = gossip_jabber_presence_state_to_str (chatrooms->presence);
 
 	if (show) {
                 lm_message_node_add_child (m->node, "show", show);
 	}
 
-        id_str = g_strdup_printf (JOIN_MSG_ID_PREFIX "%d", room->id);
+        id_str = g_strdup_printf (JOIN_MSG_ID_PREFIX "%d", id);
         lm_message_node_set_attribute (m->node, "id", id_str);
         g_free (id_str);
 
-        data = g_new0 (AsyncCallbackData, 1);
-
-	data->chatrooms = chatrooms;
-        data->callback = callback;
-        data->user_data = user_data;
-
+	/* send message */
+	room->connection = lm_connection_ref (chatrooms->connection);
         room->join_handler = lm_message_handler_new ((LmHandleMessageFunction) 
 						     jabber_chatrooms_join_cb,
-						     data, g_free);
+						     room, NULL);
 	
 	lm_connection_register_message_handler (chatrooms->connection,
 						room->join_handler, 
@@ -467,131 +651,123 @@ gossip_jabber_chatrooms_join (GossipJabberChatrooms *chatrooms,
 
 	lm_message_unref (m);
 
-	return room->id;
+	return id;
 }
 
 static LmHandlerResult
-jabber_chatrooms_join_cb (LmMessageHandler  *handler,
-                          LmConnection      *connection,
-                          LmMessage         *m,
-                          AsyncCallbackData *data)
+jabber_chatrooms_join_cb (LmMessageHandler *handler,
+                          LmConnection     *connection,
+                          LmMessage        *m,
+                          JabberChatroom   *room)
 {
-	GossipJabberChatrooms    *chatrooms;
-	GossipChatroomJoinResult  result = GOSSIP_CHATROOM_JOIN_OK;
-	GossipChatroomJoinCb      callback;	
-	JabberChatroom           *room;
+ 	GossipJabberChatrooms    *chatrooms; 
+	GossipChatroomJoinResult  result;
+	GossipChatroomStatus      status;
+	GossipChatroomId          id;
 	LmMessageSubType          type;
-	LmMessageNode            *node;
-	const gchar              *jid_str;
-	const gchar              *id_str;
-        gint                      id;
+	LmMessageNode            *node = NULL;
 
-	chatrooms = data->chatrooms;
-
-	id_str = lm_message_node_get_attribute (m->node, "id");
-	jid_str = lm_message_node_get_attribute (m->node, "from");
-
-	room = NULL;
-	/* some servers don't use the ID in the response, so this
-	   fails, so we check the JID too */
-	if (id_str) {
-		gchar *int_str;
-
-		if (strlen (id_str) < strlen (JOIN_MSG_ID_PREFIX)) {
-			g_warning ("Wrong id string when joining chatroom");
-			return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
-		}
-
-		int_str = g_strdup (id_str + strlen (JOIN_MSG_ID_PREFIX));
-		id = atoi (int_str);
-		g_free (int_str);
-
-		room = g_hash_table_lookup (chatrooms->room_id_hash, 
-					    GINT_TO_POINTER (id));
-	} else if (jid_str) {
-		GossipJID *jid;
-
-		jid = gossip_jid_new (jid_str);
-		room = g_hash_table_lookup (chatrooms->room_jid_hash, jid);
-		gossip_jid_unref (jid);
-	} 
-
-	if (!room) {
-		/* can't do much else, but should we respond to async
-		   callback with error? */
+	if (!room || !room->join_handler) {
  		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS; 
 	}
 
-	lm_connection_unregister_message_handler (connection, 
-						  handler, 
-						  LM_MESSAGE_TYPE_PRESENCE);
+	chatrooms = room->chatrooms;
+
+	/* clean up the join timeout */
+	if (room->timeout_id) {
+		g_source_remove (room->timeout_id);
+		room->timeout_id = 0;
+	}
+
+	/* clean up handler */
+	if (room->join_handler) {
+		lm_message_handler_unref (room->join_handler);
+		room->join_handler = NULL;
+	}
+
+	/* get room id */
+	id = jabber_chatrooms_chatroom_get_id (room);
 
         /* check for error */
 	type = lm_message_get_sub_type (m);
-	
 	if (type == LM_MESSAGE_SUB_TYPE_ERROR) {
 		node = lm_message_node_get_child (m->node, "error");
-		if (node) {
-			const gchar *str;
-			gint         code;
-
-			str = lm_message_node_get_attribute (node, "code");
-			code = str ? atoi (str) : 0;
+	}
+	
+	if (node) {
+		const gchar *str;
+		gint         code;
+		
+		str = lm_message_node_get_attribute (node, "code");
+		code = str ? atoi (str) : 0;
+		
+		switch (code) {
+		case 409: {
+			/* conflicting nickname */
+			DEBUG_MSG (("ProtocolChatrooms: ID[%d] Conflicting nickname", id));
+			result = GOSSIP_CHATROOM_JOIN_NICK_IN_USE;
+			break;
+		}
 			
-			switch (code) {
-			case 409: {
-				/* conflicting nickname */
-				d(g_print ("Chatrooms: Conflicting nickname\n"));
-				result = GOSSIP_CHATROOM_JOIN_NICK_IN_USE;
-				break;
-			}
-
-			case 502: {
-				/* unresolved hostname */
-				d(g_print ("Chatrooms: Unable to resolve hostname\n"));
-				result = GOSSIP_CHATROOM_JOIN_UNKNOWN_HOST;
-				break;
-			}
-
-			case 504: {
-				/* remote server timeout */
-				d(g_print ("Chatrooms: Join timed out\n"));
-				result = GOSSIP_CHATROOM_JOIN_TIMED_OUT;
-				break;
-			}
-
-			default:
-				d(g_print ("Chatrooms: Unhandled presence error:%d\n", code));
-				result = GOSSIP_CHATROOM_JOIN_UNKNOWN_ERROR;
-				break;
-			}
+		case 502: {
+			/* unresolved hostname */
+			DEBUG_MSG (("ProtocolChatrooms: ID[%d] Unable to resolve hostname", id));
+			result = GOSSIP_CHATROOM_JOIN_UNKNOWN_HOST;
+			break;
+		}
+			
+		case 504: {
+			/* remote server timeout */
+			DEBUG_MSG (("ProtocolChatrooms: ID[%d] Join timed out", id));
+			result = GOSSIP_CHATROOM_JOIN_TIMED_OUT;
+			break;
+		}
+			
+		default:
+			DEBUG_MSG (("ProtocolChatrooms: ID[%d] Unhandled presence error:%d", id, code));
+			result = GOSSIP_CHATROOM_JOIN_UNKNOWN_ERROR;
+			break;
 		}
-	} 
 
-	callback = data->callback;
-	(callback) (GOSSIP_CHATROOM_PROVIDER (chatrooms->jabber),
-		    result,
-		    room->id,
-		    data->user_data); 
+		/* set room state */
+		status = GOSSIP_CHATROOM_STATUS_ERROR;
+	} else {
+		result = GOSSIP_CHATROOM_JOIN_OK;
+		status = GOSSIP_CHATROOM_STATUS_ACTIVE;
+	}
 
+	gossip_chatroom_set_status (room->chatroom, status);
+	gossip_chatroom_set_last_error (room->chatroom, 
+					gossip_chatroom_provider_join_result_as_str (result));
+
+	if (room->callback != NULL) {
+		DEBUG_MSG (("ProtocolChatrooms: ID[%d] Calling back...", id));
+		(room->callback) (GOSSIP_CHATROOM_PROVIDER (chatrooms->jabber),
+				  result, id, room->user_data);
+	}
+
+	/* clean up callback data */
+	room->callback = NULL;
+	room->user_data = NULL;
+
+	/* articulate own contact presence so we appear in group chat */
 	if (result == GOSSIP_CHATROOM_JOIN_OK) {
-		gossip_contact_add_presence (room->own_contact, chatrooms->presence);
+		gossip_contact_add_presence (room->own_contact, 
+					     chatrooms->presence);
 		
 		g_signal_emit_by_name (chatrooms->jabber,
 				       "chatroom-contact-joined",
-				       room->id,
+				       id,
 				       room->own_contact);
-
 	} else {
 		/* clean up */
 		g_hash_table_remove (chatrooms->room_id_hash, 
-				     GINT_TO_POINTER (room->id));
-		g_hash_table_remove (chatrooms->room_jid_hash, room->jid);
-		jabber_chatrooms_chatroom_free (room);
-		
+				     GINT_TO_POINTER (id));
+		g_hash_table_remove (chatrooms->room_jid_hash, 
+				     room->jid);
 	}
 
-        return LM_HANDLER_RESULT_REMOVE_MESSAGE;
+        return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 }
 
 void
@@ -600,19 +776,45 @@ gossip_jabber_chatrooms_cancel (GossipJabberChatrooms *chatrooms,
 {
 	JabberChatroom *room;
 
-	room = (JabberChatroom *) g_hash_table_lookup (chatrooms->room_id_hash, 
-						       GINT_TO_POINTER (id));
+	g_return_if_fail (chatrooms != NULL);
+
+	room = g_hash_table_lookup (chatrooms->room_id_hash, 
+				    GINT_TO_POINTER (id));
 	if (!room) {
-		g_warning ("Unknown chatroom id: %d", id);
+		g_warning ("ProtocolChatrooms: Unknown chatroom id: %d\n", id);
 		return;
 	}
 
-	d(g_print ("Chatrooms: ID[%d] Cancel joining room\n", id));
+	DEBUG_MSG (("ProtocolChatrooms: ID[%d] Cancel joining room", id));
+
+	if (room->timeout_id) {
+		g_source_remove (room->timeout_id);
+		room->timeout_id = 0;
+	}
+
+	if (room->join_handler) {
+		lm_message_handler_unref (room->join_handler);
+		room->join_handler = NULL;
+	}
+
+	
+	gossip_chatroom_set_status (room->chatroom, GOSSIP_CHATROOM_STATUS_INACTIVE);
+	gossip_chatroom_set_last_error (room->chatroom, NULL);
+
+	if (room->callback != NULL) {
+		DEBUG_MSG (("ProtocolChatrooms: ID[%d] Calling back...", id));
+		(room->callback) (GOSSIP_CHATROOM_PROVIDER (chatrooms->jabber),
+				  GOSSIP_CHATROOM_JOIN_CANCELED, id, room->user_data);
+	}
+
+	/* clean up callback data */
+	room->callback = NULL;
+	room->user_data = NULL;
 
 	g_hash_table_remove (chatrooms->room_id_hash, 
-			     GINT_TO_POINTER (room->id));
-	g_hash_table_remove (chatrooms->room_jid_hash, room->jid);
-	jabber_chatrooms_chatroom_free (room);
+			     GINT_TO_POINTER (id));
+	g_hash_table_remove (chatrooms->room_jid_hash, 
+			     room->jid);
 }
 
 void 
@@ -623,14 +825,17 @@ gossip_jabber_chatrooms_send (GossipJabberChatrooms *chatrooms,
 	LmMessage      *m;
 	JabberChatroom *room;
 
-	room = (JabberChatroom *) g_hash_table_lookup (chatrooms->room_id_hash, 
+	g_return_if_fail (chatrooms != NULL);
+	g_return_if_fail (message != NULL);
+
+	room = (JabberChatroom*) g_hash_table_lookup (chatrooms->room_id_hash, 
 						       GINT_TO_POINTER (id));
 	if (!room) {
-		g_warning ("Unknown chatroom id: %d", id);
+		g_warning ("ProtocolChatrooms: Unknown chatroom id: %d", id);
 		return;
 	}
 	
-	d(g_print ("Chatrooms: ID[%d] Send message\n", id));
+	DEBUG_MSG (("ProtocolChatrooms: ID[%d] Send message", id));
 	
 	m = lm_message_new_with_sub_type (gossip_jid_get_without_resource (room->jid),
 					  LM_MESSAGE_TYPE_MESSAGE,
@@ -642,22 +847,26 @@ gossip_jabber_chatrooms_send (GossipJabberChatrooms *chatrooms,
 }
 
 void
-gossip_jabber_chatrooms_set_title (GossipJabberChatrooms *chatrooms,
-				   GossipChatroomId       id,
-				   const gchar           *new_title)
+gossip_jabber_chatrooms_change_topic (GossipJabberChatrooms *chatrooms,
+				      GossipChatroomId       id,
+				      const gchar           *new_topic)
 {
 	JabberChatroom *room;
 	const gchar    *without_resource;
 	LmMessage      *m;
 
-	room = (JabberChatroom *) g_hash_table_lookup (chatrooms->room_id_hash, 
+	g_return_if_fail (chatrooms != NULL);
+	g_return_if_fail (new_topic != NULL);
+
+	room = (JabberChatroom*) g_hash_table_lookup (chatrooms->room_id_hash, 
 						       GINT_TO_POINTER (id));
 	if (!room) {
-		g_warning ("Unknown chatroom id: %d", id);
+		g_warning ("ProtocolChatrooms: Unknown chatroom id: %d", id);
 		return;
 	}
-	d(g_print ("Chatrooms: ID[%d] Set chatroom title to:'%s'\n", 
-		   id, new_title));
+
+	DEBUG_MSG (("ProtocolChatrooms: ID[%d] Change topic to:'%s'", 
+		   id, new_topic));
 	
 	without_resource = gossip_jid_get_without_resource (room->jid);
 	
@@ -665,7 +874,7 @@ gossip_jabber_chatrooms_set_title (GossipJabberChatrooms *chatrooms,
 					  LM_MESSAGE_TYPE_MESSAGE,
 					  LM_MESSAGE_SUB_TYPE_GROUPCHAT);
 	
-	lm_message_node_add_child (m->node, "subject", new_title);
+	lm_message_node_add_child (m->node, "subject", new_topic);
 
 	lm_connection_send (chatrooms->connection, m, NULL);
 	lm_message_unref (m);
@@ -679,14 +888,17 @@ gossip_jabber_chatrooms_change_nick (GossipJabberChatrooms *chatrooms,
 	LmMessage      *m;
 	JabberChatroom *room;
 
-	room = (JabberChatroom *) g_hash_table_lookup (chatrooms->room_id_hash, 
+	g_return_if_fail (chatrooms != NULL);
+	g_return_if_fail (new_nick != NULL);
+
+	room = (JabberChatroom*) g_hash_table_lookup (chatrooms->room_id_hash, 
 						       GINT_TO_POINTER (id));
 	if (!room) {
-		g_warning ("Unknown chatroom id: %d", id);
+		g_warning ("ProtocolChatrooms: Unknown chatroom id: %d", id);
 		return;
 	}
 
-	d(g_print ("Chatrooms: ID[%d] Change chatroom nick to:'%s'\n", 
+	DEBUG_MSG (("ProtocolChatrooms: ID[%d] Change chatroom nick to:'%s'", 
 		   id, new_nick));
 
 	gossip_jid_set_resource (room->jid, new_nick);
@@ -705,14 +917,17 @@ gossip_jabber_chatrooms_leave (GossipJabberChatrooms *chatrooms,
 	LmMessage      *m;
 	JabberChatroom *room;
 
-	room = (JabberChatroom *) g_hash_table_lookup (chatrooms->room_id_hash, 
+	g_return_if_fail (chatrooms != NULL);
+
+	room = (JabberChatroom*) g_hash_table_lookup (chatrooms->room_id_hash, 
 						       GINT_TO_POINTER (id));
 	if (!room) {
-		g_warning ("Unknown chatroom id: %d", id);
+		g_warning ("ProtocolChatrooms: Unknown chatroom id: %d", id);
 		return;
 	}
 
-	d(g_print ("Chatrooms: ID[%d] Leaving room\n", id));
+	gossip_chatroom_set_status (room->chatroom, GOSSIP_CHATROOM_STATUS_INACTIVE);
+	gossip_chatroom_set_last_error (room->chatroom, NULL);
 
 	m = lm_message_new_with_sub_type (gossip_jid_get_full (room->jid),
 					  LM_MESSAGE_TYPE_PRESENCE,
@@ -721,10 +936,15 @@ gossip_jabber_chatrooms_leave (GossipJabberChatrooms *chatrooms,
 	lm_connection_send (chatrooms->connection, m, NULL);
 	lm_message_unref (m);
 
-	g_hash_table_remove (chatrooms->room_id_hash, 
-			     GINT_TO_POINTER (room->id));
-	g_hash_table_remove (chatrooms->room_jid_hash, room->jid);
-	jabber_chatrooms_chatroom_free (room);
+	g_hash_table_remove (chatrooms->room_id_hash,
+			     GINT_TO_POINTER (id));
+	g_hash_table_remove (chatrooms->room_jid_hash,
+			     room->jid);
+
+	DEBUG_MSG (("ProtocolChatrooms: ID[%d] Leaving room, ref count is %d", 
+		   id, room->ref_count - 1));
+	
+	jabber_chatrooms_chatroom_unref (room);
 }
 
 const gchar *
@@ -733,14 +953,16 @@ gossip_jabber_chatrooms_get_room_name (GossipJabberChatrooms *chatrooms,
 {
 	JabberChatroom *room;
 
-	room = (JabberChatroom *) g_hash_table_lookup (chatrooms->room_id_hash,
-						       GINT_TO_POINTER (id));
+	g_return_val_if_fail (chatrooms != NULL, NULL);
+
+	room = (JabberChatroom*) g_hash_table_lookup (chatrooms->room_id_hash,
+						      GINT_TO_POINTER (id));
 	if (!room) {
-		g_warning ("Unknown chatroom id: %d", id);
+		g_warning ("ProtocolChatrooms: Unknown chatroom id: %d", id);
 		return NULL;
 	}
 
-	return room->name;
+	return gossip_chatroom_get_name (room->chatroom);
 }
 
 void 
@@ -753,14 +975,17 @@ gossip_jabber_chatrooms_invite (GossipJabberChatrooms *chatrooms,
 	LmMessageNode  *n;
 	JabberChatroom *room;
 
+	g_return_if_fail (chatrooms != NULL);
+	g_return_if_fail (contact_id != NULL);
+
 	room = (JabberChatroom *) g_hash_table_lookup (chatrooms->room_id_hash, 
 						       GINT_TO_POINTER (id));
 	if (!room) {
-		g_warning ("Unknown chatroom id: %d", id);
+		g_warning ("ProtocolChatrooms: Unknown chatroom id: %d", id);
 		return;
 	}
 
-	d(g_print ("Chatrooms: ID[%d] Inviting contact:'%s'\n", 
+	DEBUG_MSG (("ProtocolChatrooms: ID[%d] Inviting contact:'%s'", 
 		   id, contact_id));
 
 	m = lm_message_new (contact_id,
@@ -786,11 +1011,13 @@ gossip_jabber_chatrooms_invite_accept (GossipJabberChatrooms *chatrooms,
 				       const gchar           *nickname,
 				       const gchar           *invite_id)
 {
-	gchar       *room = NULL;
-	const gchar *server;
+	GossipChatroom *chatroom;
+	gchar          *room = NULL;
+	const gchar    *server;
 
 	g_return_if_fail (invite_id != NULL);
 	g_return_if_fail (callback != NULL);
+
 	server = strstr (invite_id, "@");
 
 	g_return_if_fail (server != NULL);
@@ -801,14 +1028,20 @@ gossip_jabber_chatrooms_invite_accept (GossipJabberChatrooms *chatrooms,
 		server++;
 	}
 
+	chatroom = g_object_new (GOSSIP_TYPE_CHATROOM, 
+				 "type", GOSSIP_CHATROOM_TYPE_NORMAL,
+				 "server", server,
+				 "name", room,
+				 "room", room,
+				 "nick", nickname,
+				 NULL);
+
 	gossip_jabber_chatrooms_join (chatrooms,
-				      room,
-				      server,
-				      nickname,
-				      NULL,
+				      chatroom,
 				      callback,
 				      NULL);
 
+	g_object_unref (chatroom);
 	g_free (room);
 }
 
@@ -817,6 +1050,8 @@ gossip_jabber_chatrooms_get_rooms (GossipJabberChatrooms *chatrooms)
 {
 	GList *list = NULL;
 
+	g_return_val_if_fail (chatrooms != NULL, NULL);
+
 	g_hash_table_foreach (chatrooms->room_id_hash, 
 			      (GHFunc) jabber_chatrooms_get_rooms_foreach,
 			      &list);	
@@ -836,6 +1071,8 @@ void
 gossip_jabber_chatrooms_set_presence (GossipJabberChatrooms  *chatrooms,
 				      GossipPresence         *presence)
 {
+	g_return_if_fail (chatrooms != NULL);
+	
 	if (chatrooms->presence) {
 		g_object_unref (chatrooms->presence);
 	}
@@ -861,7 +1098,7 @@ jabber_chatrooms_set_presence_foreach (gpointer               key,
 	connection = chatrooms->connection;
 	presence = chatrooms->presence;
 
-	show   = gossip_jabber_helper_presence_state_to_string (presence);
+	show   = gossip_jabber_presence_state_to_str (presence);
 	status = gossip_presence_get_status (presence);
 
 	m = lm_message_new_with_sub_type (gossip_jid_get_full (room->jid),
